import numpy as np
import copy as cp
from learners.discreteMDPs.utils import allmax, categorical_sample
from learners.discreteMDPs.AgentInterface import Agent


class UCRL2_5(Agent):
    def __init__(self, env, delta, eta=1.12):
        Agent.__init__(self, env, name="UCRL2_5")
        """
        Slightly modified UCRL2 based on
        "Jaksch, Thomas, Ronald Ortner, and Peter Auer.
        'Near-optimal regret bounds for reinforcement learning'
        Journal of Machine Learning Research 11.Apr (2010): 1563-1600."

        delta: float
        Confidence level in (0,1).
        """
        self.t = 1
        self.delta = delta
        self.eta = eta

        self.observations = [[], [], []]  # list of the observed (states, actions, rewards) ordered by time
        self.vk = np.zeros((self.nS, self.nA))  # the state-action count for the current episode k
        self.Nk = np.zeros((self.nS, self.nA))  # the state-action count prior to episode k

        self.r_distances = np.zeros((self.nS, self.nA))
        self.p_distances = np.zeros((self.nS, self.nA))
        self.Pk = np.zeros((self.nS, self.nA, self.nS))
        self.Rk = np.zeros((self.nS, self.nA))
        self.Rk_old = np.zeros((self.nS, self.nA))
        self.M2 = np.zeros((self.nS, self.nA))

        self.u = np.zeros(self.nS)
        self.span = []
        self.policy = np.zeros((self.nS, self.nA))  # policy, seen as a stochastic policy here.
        for s in range(self.nS):
            for a in range(self.nA):
                self.policy[s, a] = 1. / self.nA

    def updateN(self):
        """Auxiliary function to update N the current state-action count.
        """
        for s in range(self.nS):
            for a in range(self.nA):
                self.Nk[s, a] += self.vk[s, a]

    def updateR(self):
        """Auxiliary function to update R the accumulated reward.
        """
        self.Rk_old[self.observations[0][-2], self.observations[1][-1]] = self.Rk[self.observations[0][-2], self.observations[1][-1]]
        self.Rk[self.observations[0][-2], self.observations[1][-1]] += self.observations[2][-1]

    def updateM2(self):
        """Auxiliary function to update M2 the accumulated squared reward difference from the mean reward.
        Sequential updates are performed using Welford's algorithm.
        The sample variance for (state, action) = (s, a) is M2[s, a] / Nk[s, a].
        """
        s, a = self.observations[0][-2], self.observations[1][-1]
        r = self.observations[2][-1]
        self.M2[s, a] += (r - self.Rk[s, a]) * (r - self.Rk_old[s, a])

    def updateP(self):
        """Auxiliary function to update P the transitions count.
        """
        self.Pk[self.observations[0][-2], self.observations[1][-1], self.observations[0][-1]] += 1

    def distances(self):
        """Auxiliary function updating the values of r_distances
        and p_distances (i.e. the confidence bounds used to build the set of
        plausible MDPs).
        """
        for s in range(self.nS):
            for a in range(self.nA):
                self.r_distances[s, a] = # TODO
                self.p_distances[s, a] = np.sqrt((14 * self.nS * np.log(2 * self.nA * self.t / self.delta))
                                                 / (max([1, self.Nk[s, a]])))

    def max_proba(self, p_estimate, sorted_indices, s, a):
        """Computing the maximum proba in the Extended Value Iteration
        for given state s and action a.
        """
        min1 = min([1, p_estimate[s, a, sorted_indices[-1]] + (self.p_distances[s, a] / 2)])
        max_p = np.zeros(self.nS)
        if min1 == 1:
            max_p[sorted_indices[-1]] = 1
        else:
            max_p = cp.deepcopy(p_estimate[s, a])
            max_p[sorted_indices[-1]] += self.p_distances[s, a] / 2
            l = 0
            while sum(max_p) > 1:
                max_p[sorted_indices[l]] = max([0, 1 - sum(max_p) + max_p[sorted_indices[l]]])  # Error?
                l += 1
        return max_p

    def max_r(self, r_estimate, s, a):
        """Computing the maximum reward in the Extended Value Iteration
        for given state s and action a.
        """
        # TODO

    def EVI(self, r_estimate, p_estimate, epsilon=0.01, max_iter=1000):
        """The Extend Value Iteration algorithm (approximated with precision
        epsilon), in parallel policy updated with the greedy one.
        """
        # This initialization sligthly speeds up convergence
        u0 = self.u - min(self.u)
        u1 = np.zeros(self.nS)
        sorted_indices = np.arange(self.nS)
        niter = 0
        while True:
            niter += 1
            for s in range(self.nS):
                temp = np.zeros(self.nA)
                for a in range(self.nA):
                    max_p = self.max_proba(p_estimate, sorted_indices, s, a)
                    max_r = self.max_r(r_estimate, s, a)
                    temp[a] = min(1, max_r) + sum(
                        [u * p for (u, p) in zip(u0, max_p)])
                # This implements a tie-breaking rule by choosing:  Uniform(Argmmin(Nk))
                (u1[s], arg) = allmax(temp)
                nn = [-self.Nk[s, a] for a in arg]
                (nmax, arg2) = allmax(nn)
                choice = [arg[a] for a in arg2]
                self.policy[s] = [1. / len(choice) if x in choice else 0 for x in range(self.nA)]

            diff = [abs(x - y) for (x, y) in zip(u1, u0)]
            if (max(diff) - min(diff)) < epsilon:
                self.u = u1 - min(u1)
                break
            else:
                u0 = u1 - min(u1)
                u1 = np.zeros(self.nS)
                sorted_indices = np.argsort(u0)
            if niter > max_iter:
                self.u = u1 - min(u1)
                print("No convergence in EVI")
                break

    def new_episode(self):
        """To start a new episode (init var, computes estmates and run EVI).
        """
        self.updateN()
        self.vk = np.zeros((self.nS, self.nA))
        r_estimate = np.zeros((self.nS, self.nA))
        p_estimate = np.zeros((self.nS, self.nA, self.nS))
        for s in range(self.nS):
            for a in range(self.nA):
                div = max([1, self.Nk[s, a]])
                r_estimate[s, a] = self.Rk[s, a] / div
                for next_s in range(self.nS):
                    p_estimate[s, a, next_s] = self.Pk[s, a, next_s] / div
        self.distances()
        self.EVI(r_estimate, p_estimate, epsilon=1. / max(1, self.t))

    def reset(self, inistate):
        """To reinitialize the learner with a given initial state inistate.
        """
        self.t = 1
        self.observations = [[inistate], [], []]
        self.vk = np.zeros((self.nS, self.nA))
        self.Nk = np.zeros((self.nS, self.nA))
        self.u = np.zeros(self.nS)
        self.Pk = np.zeros((self.nS, self.nA, self.nS))
        self.Rk = np.zeros((self.nS, self.nA))
        self.M2 = np.zeros((self.nS, self.nA))
        self.span = [0]
        for s in range(self.nS):
            for a in range(self.nA):
                self.policy[s, a] = 1. / self.nA
        self.new_episode()

    def play(self, state):
        """To chose an action for a given state (and start a new episode if
        necessary -> stopping criterion defined here).
        """
        action = categorical_sample([self.policy[state, a] for a in range(self.nA)], np.random)
        if self.vk[state, action] >= max([1, self.Nk[state, action]]):  # Stoppping criterion
            self.new_episode()
            action = categorical_sample([self.policy[state, a] for a in range(self.nA)], np.random)
        return action

    def update(self, state, action, reward, observation):
        """To update the learner after one step of the current policy.
        """
        self.vk[state, action] += 1
        self.observations[0].append(observation)
        self.observations[1].append(action)
        self.observations[2].append(reward)
        self.updateP()
        self.updateR()
        self.updateM2()
        self.t += 1
